/**

    Copyright 2007 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package it.eng.spagosoa.smx.components.jdbcstore;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.ArrayList;

import org.apache.commons.dbutils.QueryRunner;
import org.apache.commons.dbutils.ResultSetHandler;

/**
 * Store Procedure runner. This runner use a simplistic assumption where
 * the store procedure is declared with all out parameters first and input
 * ones after.
 * <pre>
 * (i.e.: {&lt;out(1)> = call &lt;functionname>(&lt;out(2)>, ..., &lt;out(n)>, &lt;in(1)>, ..., &lt;in(m)> ) })
 * </pre>
 * 
 * @author buso
 *
 */

/*
 * Propose this code to DBUtils maintainer.
 */
public class StoreProcedureRunner extends QueryRunner {
	
	public Object call(String call, Object[] params, int[] outParamTypes, ResultSetHandler rsh) throws SQLException {
		Connection conn = prepareConnection();
		
        try {
            return call(conn, call, params, outParamTypes, rsh);
        } finally {
            close(conn);
        }		
	}

	public Object call(Connection conn, String call,
		Object[] params, int[] outParamTypes, ResultSetHandler rsh) throws SQLException {
		
        CallableStatement stmt = null;
        ResultSet rs = null;
        Object result = null;

        try {
            stmt = conn.prepareCall(call);
            int offSet = outParamTypes != null? outParamTypes.length : 0;
            this.fillStatement(stmt, params, offSet);
            registerOutParameters(stmt, outParamTypes);
            stmt.execute();
            return retrieveOutValues(outParamTypes, stmt, rsh);
        } catch (SQLException e) {
            this.rethrow(e, call, params);

        } finally {
            try {
                close(rs);
            } finally {
                close(stmt);
            }
        }

        return result;		
	}

    protected Object[] retrieveOutValues(int[] out, CallableStatement stmt, ResultSetHandler rsh)
    	throws SQLException {
    	
        if (null == out || out.length == 0)
        {
            return new Object[0];
        }
        
        ArrayList<Object> list = new ArrayList<Object>();
        for (int i = 0; i < out.length; i++)
        {
        	Object elem = stmt.getObject(i + 1);
        	if (elem instanceof ResultSet) {
        		list.add(rsh.handle((ResultSet)elem));
        	} else {
        		list.add(elem);
        	}
        }
        return list.toArray();
    }
    
	/**
	 * We are assuming input parameters are passed after all out ones.
	 * 
	 * @param offSet. Starting input params position in store procedure declaration
	 */
    protected void fillStatement(PreparedStatement stmt, Object[] params, int offSet)
	    throws SQLException {
	
	    if (params == null) {
	        return;
	    }
	    for (int i = 0; i < params.length; i++) {
	        if (params[i] != null) {
	            stmt.setObject(i + offSet + 1, params[i]);
	        } else {
	            // VARCHAR works with many drivers regardless
	            // of the actual column type.  Oddly, NULL and 
	            // OTHER don't work with Oracle's drivers.
	            stmt.setNull(i + offSet + 1, Types.VARCHAR);
	        }
	    }
	}	
	protected void registerOutParameters(CallableStatement stmt,
		int[] outParams) throws SQLException {
		
		for (int outIdx = 0; outIdx < outParams.length; outIdx++) {
			stmt.registerOutParameter(outIdx+1, outParams[outIdx]);
		}
	}
}
