package it.eng.spagosoa.smx.components.restart;

import it.eng.spagosoa.metadb.dbutils.HibernateUtil;
import it.eng.spagosoa.metadb.model.JournalInstanceService;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.servicemix.id.IdGenerator;
import org.apache.servicemix.jbi.container.JBIContainer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.dom4j.Document;
import org.dom4j.XPath;
import org.dom4j.tree.DefaultElement;
import org.dom4j.xpath.DefaultXPath;
import org.hibernate.Session;

/**
 * Manages the restart of a list of Instance SA, without the possibility of changing the messages
 * The InstanceSA, in the list, have to be faulted.
 *
 */
public class MultipleInstanceRestart extends RestartBase {

	// Massive Instance Restart expressions
	private static final String ID_LIST_INSTANCES_XPATH_EXPRESSION = "//restart:restartMultipleInstances/restart:idInstanceSA";
	
	public MultipleInstanceRestart(JBIContainer container, Map<String, String> namespaceURI, IdGenerator idGenerator, RestartComponent restartComponent) {
		super(container, namespaceURI, idGenerator, restartComponent);
	}


	/**
	 * Perform a massive restart.
	 * @param exchange Request exchange
	 * @param document The DOM document of the message
	 * @throws Exception
	 */
	public void manageMassiveRestart(MessageExchange exchange, Document document, NormalizedMessage out) throws Exception {
		// Sample Request Message
		//			<restartMultipleInstances xmlns="urn:it:eng:spagic:restart">
		//			<idInstanceSA>1</idInstanceSA>
		//			<idInstanceSA>2</idInstanceSA>
		//			<idInstanceSA>3</idInstanceSA>
		//			</restartMultipleInstances>
		
		long[] instancesSAToRestart = retrieveInstancesSAToRestart(document);
		long[] instancesSARestarted = new long[instancesSAToRestart.length];
		String[] errorMessage = new String[instancesSAToRestart.length];
		
		Session aSession = null;
		try {
			aSession = HibernateUtil.getSessionFactory().openSession();
			
			// Prepare an empty list of modified messages: in massive restart
			// it isnt possible
			List<ModifiedMessage> modifiedMessages = new ArrayList<ModifiedMessage>(0);
			
			for (int i = 0; i < instancesSAToRestart.length; i++) {
				long idInstanceSA = instancesSAToRestart[i];
				instancesSARestarted[i] = -1;
				errorMessage[i] = null;
				
				try {
					PrepareRestartResult prepareResult = prepareInstanceSARestart(aSession, idInstanceSA);
					
					// Check for errors
					if (prepareResult.errorMessage != null) {
						// Store error message
						errorMessage[i] = prepareResult.errorMessage;
						
						// Proceed with next SA instance
						continue;
					}
					
					String newCorrelationId = prepareResult.instanceSADuplicated.getIdCorrelation();
					for (JournalInstanceService jis : prepareResult.journalInstanceServices) {
						if (!restartSingleJIS(jis, newCorrelationId, modifiedMessages)) {
							// Store error message
							errorMessage[i] = "Exchange found on database with wrong state. JIS id: " + jis.getIdJournal();
							
							// Proceed with next SA instance
							break;
						}
					}
					instancesSARestarted[i] = prepareResult.instanceSADuplicated.getIdInstanceSa();
				}
				catch (Exception ex) {
					// Manage error for this SA instance
					// Store error message
					errorMessage[i] = ex.getMessage();
				}
			}
			
			// Sample Response Message
			//		   <restartMultipleInstancesResponse xmlns="urn:it:eng:spagic:restart">
			//		      <restartMultipleInstancesReturn>
			//		         <restartedInstanceId>1</restartedInstanceId>
			//		         <errorMessage>no error</errorMessage>
			//		      </restartMultipleInstancesReturn>
			//		      <restartMultipleInstancesReturn>
			//		         <restartedInstanceId>2</restartedInstanceId>
			//		         <errorMessage>simple error</errorMessage>
			//		      </restartMultipleInstancesReturn>
			//		      <restartMultipleInstancesReturn>
			//		         <restartedInstanceId>3</restartedInstanceId>
			//		         <errorMessage>very complex error</errorMessage>
			//		      </restartMultipleInstancesReturn>
			//		   </restartMultipleInstancesResponse>

			
			// Send response
			StringBuffer responseBuffer = new StringBuffer();
			responseBuffer.append("<restartMultipleInstancesResponse xmlns=\"urn:it:eng:spagic:restart\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">");
			
			for (int i = 0; i < instancesSARestarted.length; i++) {
				responseBuffer.append("<restartMultipleInstancesReturn>");
				
				responseBuffer.append("<restartedInstanceId>");
				responseBuffer.append(instancesSARestarted[i]);
				responseBuffer.append("</restartedInstanceId>");
				if (errorMessage[i] != null) {
					responseBuffer.append("<errorMessage><![CDATA[");
					responseBuffer.append(errorMessage[i]);
					responseBuffer.append("]]></errorMessage>");
				} else {
					responseBuffer.append("<errorMessage xsi:nil=\"true\" />");
				}
				
				responseBuffer.append("</restartMultipleInstancesReturn>");
			}
			
			responseBuffer.append("</restartMultipleInstancesResponse>");
			out.setContent(new StringSource(responseBuffer.toString()));
		}
		finally {
			if (aSession != null) {
				aSession.close();
			}
		}
	}
	
	
	/**
	 * Extract from the DOM the id list of the InstanceSA to restart
	 * @param document DOM to parse
	 * @return List of Id of the InstanceSA
	 */
	private long[] retrieveInstancesSAToRestart(Document document) {
		XPath instancesSAToRestartXPathExpression = new DefaultXPath(ID_LIST_INSTANCES_XPATH_EXPRESSION);
		instancesSAToRestartXPathExpression.setNamespaceURIs(getNamespaceURI());
		List<DefaultElement> nodes = instancesSAToRestartXPathExpression.selectNodes(document);
		long[] instancesSAToRestart = new long[nodes.size()];
		
		int i = 0;
		for (DefaultElement currentNode : nodes) {
			String idInstanceSA = currentNode.getText();
			debug("idInstanceSA = " + idInstanceSA);

			instancesSAToRestart[i++] = Long.parseLong(idInstanceSA);
		}
		return instancesSAToRestart;
	}
	
}
