package it.eng.spagosoa.smx.components.restart;

import it.eng.spagosoa.metadb.base.MepStateUtils;
import it.eng.spagosoa.metadb.dbutils.AuditDBUtils;
import it.eng.spagosoa.metadb.dbutils.HibernateUtil;
import it.eng.spagosoa.metadb.model.InstanceSa;
import it.eng.spagosoa.metadb.model.JournalInstanceService;
import it.eng.spagosoa.metadb.model.ServiceAssembly;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.servicemix.id.IdGenerator;
import org.apache.servicemix.jbi.container.JBIContainer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.dom4j.Document;
import org.hibernate.Session;

/**
 * Manage the restart of a list of Instance SA, starting from a list of messages.
 * This feature works also (or only?) with non-faulted messages.
 *
 */
public class MultipleMessageRestart extends RestartBase {

	// Multiple Message Restart expressions
	private static final String MULTIPLE_MESSAGE_MODIFIED_MESSAGES_XPATH_EXPRESSION = "//restart:restartMultipleMessages/restart:modifiedMessages/restart:modifiedMessage";

	/**
	 * Comparator used to group a list of JournalInstanceService by InstanceSa id
	 */
	public class JISComparator implements Comparator<JournalInstanceService> {

		public int compare(JournalInstanceService o1, JournalInstanceService o2) {
			if (o1.getInstanceSa().getIdInstanceSa() < o2.getInstanceSa().getIdInstanceSa()) {
				return -1;
			} else if (o1.getInstanceSa().getIdInstanceSa() == o2.getInstanceSa().getIdInstanceSa()) {
				return 0;
			}
			return 1;
		}
		
	}
	
	
	public MultipleMessageRestart(JBIContainer container, Map<String, String> namespaceURI, IdGenerator idGenerator, RestartComponent restartComponent) {
		super(container, namespaceURI, idGenerator, restartComponent);
	}

	/**
	 * Perform a massive message restart
	 * @param exchange Restart request exchange
	 * @param document DOM
	 * @param out Response Message
	 * @throws Exception
	 */
	public void manageMassiveMessagesRestart(MessageExchange exchange, Document document, NormalizedMessage out) throws Exception {
	
		// Sample request message
		//		<restartMultipleMessages xmlns="urn:it:eng:spagic:restart" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
		//		   <modifiedMessages>
		//		      <modifiedMessage>
		//		         <idJournal>123</idJournal>
		//		         <xmlMessage>
		//		         .....
		//		         </xmlMessage>
		//		      </modifiedMessage>
		//		      <modifiedMessage>
		//		         <idJournal>456</idJournal>
		//		         <xmlMessage xsi:nil="true" />
		//		         </xmlMessage>
		//		      </modifiedMessage>
		//		   </modifiedMessages>
		//		</restartMultipleMessages>
		
		
		// Retrieve all messages to restart
		List<ModifiedMessage> modifiedMessages = retrieveModifiedMessages(document, MULTIPLE_MESSAGE_MODIFIED_MESSAGES_XPATH_EXPRESSION);
		debug("Number of messages to restart: " + modifiedMessages.size());
		
		if (modifiedMessages.size() == 0) {
			// At least one message should be specified
			throw new MessagingException("No messages specified for restart");
		}
		
		
		// Group messages for InstanceSa and duplicate the InstanceSa
		PrepareRestartResult[] setupResult = retrieveInstancesSAToRestartFromMessages(document, modifiedMessages);

		// Now all instances to restart have been duplicated: we can restart the messages
		// that the user chose
		for (PrepareRestartResult prepareResult : setupResult) {
			// Skip instances with previous problems
			if (prepareResult.errorMessage != null) {
				continue;
			}
				
			String newCorrelationId = prepareResult.instanceSADuplicated.getIdCorrelation();
			for (JournalInstanceService jis : prepareResult.journalInstanceServices) {
				if (!restartSingleJIS(jis, newCorrelationId, modifiedMessages)) {
					// Store error message
					prepareResult.errorMessage = "Exchange found on database with wrong state. JIS id: " + jis.getIdJournal();
					
					// Proceed with next SA instance
					break;
				}
			}
		}
		
	
		// Sample Response Message
		// Except for the root element, the message has the same format of
		// the multiple instance restart response
		//
		//		   <restartMultipleMessagesResponse xmlns="urn:it:eng:spagic:restart">
		//		      <restartMultipleInstancesReturn>
		//		         <restartedInstanceId>1</restartedInstanceId>
		//		         <errorMessage>no error</errorMessage>
		//		      </restartMultipleInstancesReturn>
		//		      <restartMultipleInstancesReturn>
		//		         <restartedInstanceId>2</restartedInstanceId>
		//		         <errorMessage>simple error</errorMessage>
		//		      </restartMultipleInstancesReturn>
		//		      <restartMultipleInstancesReturn>
		//		         <restartedInstanceId>3</restartedInstanceId>
		//		         <errorMessage>very complex error</errorMessage>
		//		      </restartMultipleInstancesReturn>
		//		   </restartMultipleMessagesResponse>

		
		// Send response
		StringBuffer responseBuffer = new StringBuffer();
		responseBuffer.append("<restartMultipleMessagesResponse xmlns=\"urn:it:eng:spagic:restart\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">");
		
		for (PrepareRestartResult executionResult : setupResult) {
			responseBuffer.append("<restartMultipleInstancesReturn>");
			
			responseBuffer.append("<restartedInstanceId>");
			responseBuffer.append((executionResult.instanceSADuplicated != null) ? executionResult.instanceSADuplicated.getIdInstanceSa() : "-1");
			responseBuffer.append("</restartedInstanceId>");
			if (executionResult.errorMessage != null) {
				responseBuffer.append("<errorMessage><![CDATA[");
				responseBuffer.append(executionResult.errorMessage);
				responseBuffer.append("]]></errorMessage>");
			} else {
				responseBuffer.append("<errorMessage xsi:nil=\"true\" />");
			}
			
			responseBuffer.append("</restartMultipleInstancesReturn>");
		}
		
		responseBuffer.append("</restartMultipleMessagesResponse>");
		out.setContent(new StringSource(responseBuffer.toString()));

	}
	
	protected PrepareRestartResult prepareInstanceSARestart(Session aSession, List<JournalInstanceService> jisList) throws Exception {
		PrepareRestartResult result = new PrepareRestartResult();
		
		// Retrieve the instance to duplicate
		InstanceSa instanceToDuplicate = jisList.get(0).getInstanceSa();
		result.instanceSAToRestart = instanceToDuplicate;
		
		
		// Check if all exchanges are InOnly
		for (JournalInstanceService currentJIS : jisList) {
			if (MepStateUtils.isInOut(currentJIS.getMepState())) {
				result.errorMessage = "The exchange to restart is In-Out (Journal Instance Service Id: " + currentJIS.getIdJournal() + "). It's not allowed to restart In-Out exchanges.";
				log.error(result.errorMessage);
				return result;
			}
		}
		
		// Check if the version of the instanceSA to restart is equal to the latest
		// version on database, for this process
		ServiceAssembly serviceAssembly = instanceToDuplicate.getServiceAssembly();
		long saVersion = serviceAssembly.getSaVersion();
		Long latestSAVersion = AuditDBUtils.getCurrentServiceAssemblyVersion(serviceAssembly.getName(), aSession);
		
		if ((latestSAVersion == null) || (latestSAVersion.longValue() != saVersion)) {
			result.errorMessage = "The exchange to restart is not deployed anymore on ServiceManager. Version of the Service Assembly to restart: " + saVersion + ", Version of the Service Assembly deployed: " + latestSAVersion;
			log.error(result.errorMessage);
			return result;
		}
		
		
		// Retrieves all the Journal_instance_service to restart for this this Service Assembly
		result.journalInstanceServices = jisList;
		
		// Generate a new correlation id for duplicated instance
		String newCorrelationId = idGenerator.generateId();
		debug("Generated correlation id: " + newCorrelationId);

		// Duplicate the instance of the Service Assembly
		result.instanceSADuplicated = AuditDBUtils.duplicateInstanceSA(instanceToDuplicate, newCorrelationId, false, aSession);
		debug("Duplicated instance with id: " + result.instanceSAToRestart.getIdInstanceSa() + ". New duplicated id: " + result.instanceSADuplicated.getIdInstanceSa());
		
		return result;
	}

	/**
	 * Receives a group of JournalInstanceService related to the same InstanceSA and perform
	 * the setup activities before restarting these JournalInstanceService
	 * @param jisGroup Group of JournalInstanceService
	 * @param aSession Hibernate session
	 * @return PrepareRestartResult contains the id of the restarted InstanceSA or an error message
	 */
	private PrepareRestartResult prepareRestartForJISGroup(List<JournalInstanceService> jisGroup, Session aSession) {
		PrepareRestartResult prepareRestartResult = null;
		try {
			prepareRestartResult = prepareInstanceSARestart(aSession, jisGroup);
		}
		catch (Exception ex) {
			// Manage error for this SA instance
			// Store error message
			prepareRestartResult.errorMessage = ex.getMessage();
			prepareRestartResult.instanceSAToRestart = jisGroup.get(0).getInstanceSa();
			prepareRestartResult.journalInstanceServices = jisGroup;
		}
		return prepareRestartResult;
	}
	
	/**
	 * Starting from the messages, it retrieves all SA instances to restart: some messages
	 * can refer to the same SA instance
	 * @param document 
	 * @param modifiedMessages
	 * @return PrepareRestartResult[] Array of instances to restart and duplicated instances
	 */
	private PrepareRestartResult[] retrieveInstancesSAToRestartFromMessages(Document document, List<ModifiedMessage> modifiedMessages) {
		
		Session aSession = null;
		List<PrepareRestartResult> result = new ArrayList<PrepareRestartResult>();
		try {
			aSession = HibernateUtil.getSessionFactory().openSession();
		
			// Load the messages from the database
			List<JournalInstanceService> journalInstanceServices = new ArrayList<JournalInstanceService>(modifiedMessages.size());
			for (ModifiedMessage message : modifiedMessages) {
				journalInstanceServices.add(AuditDBUtils.getJournalInstanceServiceById(aSession, message.getIdJournal()));
			}
			
			// Group the messages by InstanceSa id
			Collections.sort(journalInstanceServices, new JISComparator());
			
			// Collect the messages by InstanceSa id and duplicate the InstanceSa objects
			// before restarting messages
			List<JournalInstanceService> jisGroup = new ArrayList<JournalInstanceService>();
			for (JournalInstanceService jis : journalInstanceServices) {
				if (jisGroup.size() == 0) {
					// Group empty: no element to compare with
					jisGroup.add(jis);
				} else if (jis.getInstanceSa().getIdInstanceSa() == jisGroup.get(0).getInstanceSa().getIdInstanceSa()) {
					// Same group because the InstanceSa is the same
					jisGroup.add(jis);
				} else {
					// Different group: we have to manage the current group, and then create a new one
					PrepareRestartResult prepareRestartResult = prepareRestartForJISGroup(jisGroup, aSession);
					result.add(prepareRestartResult);
					
					// Create a new group
					jisGroup = new ArrayList<JournalInstanceService>();
					jisGroup.add(jis);
				}
			}
			
			// Manage the last group
			if (jisGroup.size() != 0) {
				PrepareRestartResult prepareRestartResult = prepareRestartForJISGroup(jisGroup, aSession);
				result.add(prepareRestartResult);
			}
		}
		finally {
			if (aSession != null) {
				aSession.close();
			}
		}
		
		PrepareRestartResult[] arrayResult = new PrepareRestartResult[result.size()];
		return result.toArray(arrayResult);
	}	
	
}
