package it.eng.spagosoa.smx.components.restart;

import it.eng.spagosoa.metadb.base.MepStateConstants;
import it.eng.spagosoa.metadb.dbutils.AuditDBUtils;
import it.eng.spagosoa.metadb.model.InstanceSa;
import it.eng.spagosoa.metadb.model.JournalInstanceService;
import it.eng.spagosoa.metadb.model.ServiceAssembly;

import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.id.IdGenerator;
import org.apache.servicemix.jbi.container.JBIContainer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.dom4j.Document;
import org.dom4j.Node;
import org.dom4j.XPath;
import org.dom4j.xpath.DefaultXPath;
import org.hibernate.Session;

public class RestartBase {

	// Necessary for resending exchanges
	protected JBIContainer container = null;
	
	// Map of namespaces uri for evaluating XPath expressions
	protected Map<String, String> namespaceURI = null;
	
	// Generator for correlation ids
	protected IdGenerator idGenerator = null;
	
	// Main restart component
	protected RestartComponent restartComponent = null;
	
	protected static final Log log = LogFactory.getLog(RestartComponent.class);

	protected static final long NO_INSTANCE_SA_CREATED = -1;
	
	private static final String ID_JOURNAL_XPATH_EXPRESSION = "restart:idJournal";
	private static final String XML_MESSAGE_XPATH_EXPRESSION = "restart:xmlMessage";

	
	public RestartBase(JBIContainer container, Map<String, String> namespaceURI, IdGenerator idGenerator, RestartComponent restartComponent) {
		super();
		this.container = container;
		this.namespaceURI = namespaceURI;
		this.idGenerator = idGenerator;
		this.restartComponent = restartComponent;
	}

	/**
	 * Result of the setup activities that are performed before restarting.
	 */
	public class PrepareRestartResult {
		public InstanceSa instanceSAToRestart = null;
		public List<JournalInstanceService> journalInstanceServices = null;
		
		public InstanceSa instanceSADuplicated = null;
		public String errorMessage = null;
	}
	
	public void debug(Object o){
		if (log.isDebugEnabled()){
			log.debug(o);
		}
	}

	public void setContainer(JBIContainer container) {
		this.container = container;
	}	

	public Map<String, String> getNamespaceURI(){
		return namespaceURI;
	}

	/**
	 * Send an error notification according to the WSDL contract of the service
	 * @param restartedInstanceId
	 * @param errorMessage Error message to notify
	 * @throws RestartException
	 */
	protected void throwErrorNotification(long restartedInstanceId, String errorMessage) throws RestartException {
		throw new RestartException(restartedInstanceId, errorMessage);
	}

	/**
	 * Extract from the DOM the list of the messages modified by the user.
	 * @param document DOM to parse
	 * @return List of messages modified by the user
	 */
	protected List<ModifiedMessage> retrieveModifiedMessages(Document document, String xPathExpression) {
		// Retrieve the messages modified by the user
		XPath modifiedMessagesXPathExpression = new DefaultXPath(xPathExpression);
		modifiedMessagesXPathExpression.setNamespaceURIs(getNamespaceURI());
		List<Node> nodes = modifiedMessagesXPathExpression.selectNodes(document);
		List<ModifiedMessage> modifiedMessages = new ArrayList<ModifiedMessage>(nodes.size());
		
		// Prepare the XPath expressions to retrieve modified messages data
		XPath idJournalXPathExpression = new DefaultXPath(ID_JOURNAL_XPATH_EXPRESSION);
		idJournalXPathExpression.setNamespaceURIs(getNamespaceURI());
		XPath xmlMessageXPathExpression = new DefaultXPath(XML_MESSAGE_XPATH_EXPRESSION);
		xmlMessageXPathExpression.setNamespaceURIs(getNamespaceURI());

		for (Node node : nodes) {
			String idJournal = idJournalXPathExpression.valueOf(node);
			String modifiedXmlMessage = xmlMessageXPathExpression.valueOf(node);
			debug("idJournal = " + idJournal + ", modifiedXmlMessage = " + modifiedXmlMessage);
			
			ModifiedMessage modifiedMessage = new ModifiedMessage(Long.parseLong(idJournal), modifiedXmlMessage);
			modifiedMessages.add(modifiedMessage);
		}
		return modifiedMessages;
	}

	/**
	 * Resend a single Exchange to the JBI container.
	 * @param jis JournalInstanceService containing the exchange.
	 * @param newCorrelationId Correlation id to set in the resent exchange.
	 * @param modifiedMessages Messages modified by the user.
	 * @return True if the exchange has been resent.
	 * @throws JBIException
	 */
	protected boolean restartSingleJIS(JournalInstanceService jis, String newCorrelationId, List<ModifiedMessage> modifiedMessages) throws JBIException {
		// Rebuild the original exchange from the blob
		MessageExchange deserializedExchange = deserializeExchange(jis.getMessageExchange());
		debug("Deserialized exchange:");
		log.debug(deserializedExchange);
		
		// Check if the state of the deserialized exchange is right: only ACTIVE exchanges should
		// be saved on the database
		if (!deserializedExchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
			log.error("Exchange found on database with wrong state (should be ACTIVE). Status: " + deserializedExchange.getStatus());
			return false;
		} else {
			// Update the correlation id before resending the exchange
			deserializedExchange.setProperty(JbiConstants.CORRELATION_ID, newCorrelationId);
			
			// Check if the exchange to send has been modified by the user
			for (ModifiedMessage modifiedMessage : modifiedMessages) {
				// If the user changed the original message
				if ((modifiedMessage.getModifiedMessage() != null) &&
					(modifiedMessage.getModifiedMessage().trim().length() != 0)) {
					// If I have to resend one of the modified messages
					if (modifiedMessage.getIdJournal() == jis.getIdJournal()) {
						// Substitute message content
						NormalizedMessage originalMessage = restartComponent.getInMessage(deserializedExchange);
						originalMessage.setContent(new StringSource(modifiedMessage.getModifiedMessage()));
						break;
					}
				}
			}
			container.resendExchange(deserializedExchange);
			return true;
		}
	}

	/**
	 * Deserialize a byte stream to rebuild a Message Exchange
	 * @param buf Byte stream
	 * @return MessageExchange
	 */
	private MessageExchange deserializeExchange(byte[] buf) {
		MessageExchange exchange = null;
        
		try {
			// Deserialize from a byte array
			ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(buf));
			exchange = (MessageExchange) in.readObject();
	        in.close();
		} catch (Exception ex) {
			log.error("Error serializing exchange " + exchange, ex );
		}
        return exchange;
	}
	
	/**
	 * Perform setup activities before restarting a single process. The activities are:
	 * 	- Retrieve all exchanges to restart for current process
	 * 	- Duplicate InstanceSa and all related tables for current process
	 * @param aSession Hibernate session
	 * @param instanceSAId Process to restart
	 * @return Result of setup activity
	 * @throws Exception
	 */
	protected PrepareRestartResult prepareInstanceSARestart(Session aSession, long instanceSAId) throws Exception {
		PrepareRestartResult result = new PrepareRestartResult();
		
		// Retrieve the instance to duplicate
		InstanceSa instanceToDuplicate = AuditDBUtils.getInstanceSaById(aSession, instanceSAId);
		result.instanceSAToRestart = instanceToDuplicate;
		
		
		// Check if the version of the instanceSA to restart is equal to the latest
		// version on database, for this process
		ServiceAssembly serviceAssembly = instanceToDuplicate.getServiceAssembly();
		long saVersion = serviceAssembly.getSaVersion();
		Long latestSAVersion = AuditDBUtils.getCurrentServiceAssemblyVersion(serviceAssembly.getName(), aSession);
		
		if ((latestSAVersion == null) || (latestSAVersion.longValue() != saVersion)) {
			result.errorMessage = "The exchange to restart is not deployed anymore on ServiceManager. Version of the Service Assembly to restart: " + saVersion + ", Version of the Service Assembly deployed: " + latestSAVersion;
			log.error(result.errorMessage);
			return result;
		}
		
		
		// Retrieves all the Journal_instance_service related to this Service Assembly
		// which has a not null Message Exchange inside
		result.journalInstanceServices = retrieveJournalInstanceServicesWithExchange(instanceToDuplicate);
		
		if (result.journalInstanceServices.size() == 0) {
			result.errorMessage = "Faulted Exchanges not found on database. instanceSa Id: " + instanceSAId;
			log.error(result.errorMessage);
			return result;
		}

		// Generate a new correlation id for duplicated instance
		String newCorrelationId = idGenerator.generateId();
		debug("Generated correlation id: " + newCorrelationId);

		// Duplicate the instance of the Service Assembly
		result.instanceSADuplicated = AuditDBUtils.duplicateInstanceSA(instanceToDuplicate, newCorrelationId, aSession);
		debug("Duplicated instance id: " + instanceSAId + ". Duplicated id: " + result.instanceSADuplicated.getIdInstanceSa());
		
		return result;
	}
	
	/**
	 * Retrieve the JournalInstanceService containing the Exchange to restart.
	 * @param extractedID Message Exchange Id
	 * @return JournalInstanceService instance
	 */
	private List<JournalInstanceService> retrieveJournalInstanceServicesWithExchange(InstanceSa instanceSA) {
		List<JournalInstanceService> result = new ArrayList<JournalInstanceService>();
		
		// Search by message exchange id
		debug("Search Journal Instance Service if exist a record with idInstanceSA = "+instanceSA.getIdInstanceSa());
		Set<JournalInstanceService> journalInstanceServices = instanceSA.getJournalInstanceServices();
		
		for (Iterator it = journalInstanceServices.iterator(); it.hasNext(); ) {
			JournalInstanceService jis = (JournalInstanceService) it.next();
			// Remove all instances that doesnt have an exchange inside
			if (jis.getMessageExchange() != null) {
				int mepState = jis.getMepState().getIdState();
				if ((mepState != MepStateConstants.MEP_INONLY_FAULT) &&
					(mepState != MepStateConstants.MEP_INOUT_FAULT) &&
					(mepState != MepStateConstants.MEP_ROBUSTINONLY_STEP1_FAULT) &&
					(mepState != MepStateConstants.MEP_INOPTIONALOUT_STEP1_FAULT) &&
					(mepState != MepStateConstants.MEP_INOPTIONALOUT_STEP2_FAULT)) {
					log.error("There is a JournalInstanceService containing an exchange, but it's state is not a fault state. Id of JournalInstanceService: " + jis.getIdJournal());

//				restart INONLY exchange
//				TODO check if this is a binding choice, than we are excluding exchanges that need to restart
				} else if ((mepState != MepStateConstants.MEP_INONLY_FAULT) &&
							(mepState != MepStateConstants.MEP_ROBUSTINONLY_STEP1_FAULT)) {
					log.debug("Excluding exchange with MEP="+jis.getMepState().getStateValue()+
						". We are restarting only INONLY exchanges");
				} else {
					// Ensure that the BLOB containing the exchange is read
					jis.getMessageExchange();
					result.add(jis);
				}
			}
		}
		return result;
	}

}
