/**

    Copyright 2007 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package it.eng.spagosoa.smx.components.tcp;

import it.eng.spagosoa.smx.components.tcp.codec.TCPMsgCodecFactory;
import it.eng.spagosoa.smx.components.tcp.ssl.TCPBCSSLContextFactory;
import it.eng.spagosoa.smx.components.tcp.util.NMUtils;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.HashMap;

import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.mina.common.DefaultIoFilterChainBuilder;
import org.apache.mina.common.IoAcceptor;
import org.apache.mina.common.IoAcceptorConfig;
import org.apache.mina.common.IoSession;
import org.apache.mina.filter.LoggingFilter;
import org.apache.mina.filter.SSLFilter;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.transport.socket.nio.SocketAcceptor;
import org.apache.mina.transport.socket.nio.SocketAcceptorConfig;
import org.apache.mina.util.NewThreadExecutor;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.common.endpoints.ConsumerEndpoint;
import org.apache.servicemix.jbi.NoOutMessageAvailableException;

/**
 * @org.apache.xbean.XBean element="consumer"
 */
public class TCPConsumerEndpoint extends ConsumerEndpoint implements TCPEndpointType {
	private static final Log log = LogFactory.getLog(TCPConsumerEndpoint.class); 
	private TCPBCConfig config;

	private InetSocketAddress socketAddress;
	private TCPServerProtocolHandler serverHandler;
	private IoAcceptorConfig acceptorConfig;
	private IoAcceptor acceptor = null;
	private HashMap<String, IoSession> correlatedSessions;
	
    public void validate() throws DeploymentException {
        super.validate();
    }
    
    
    
    public String getLocationURI() {
         // TODO: return a URI that unique identify this endpoint
         return getService() +  "#" + getEndpoint(); 
    }
    
    /**
     * Start the MINA server; setting:
     * <ul>
     * 	<li>the 'codec' that recognise messages of the TCP protocol unwrapping received bytes</li>
     * 	<li>optionally add ssl support</li>
     * 	<li>the TCPServerProtocolHandler that manage messages identify by the 'codec'</li>
     * </ul>
     */
    public void start() throws Exception {
        super.start();
        
        correlatedSessions = new HashMap<String, IoSession>();
        
        // set the number of threads that process incoming data
        if (config.getConnNumber() > 0) {
        	acceptor = new SocketAcceptor(config.getConnNumber(), new NewThreadExecutor());
        } else {
        	acceptor = new SocketAcceptor();
        }
        acceptorConfig = new SocketAcceptorConfig();
        ((SocketAcceptorConfig)acceptorConfig).setReuseAddress(true);
        ((SocketAcceptorConfig)acceptorConfig).setDisconnectOnUnbind(false);
        ((SocketAcceptorConfig)acceptorConfig).setBacklog(config.getListenBacklog());
        DefaultIoFilterChainBuilder chain = acceptorConfig.getFilterChain();
        
        if (log.isDebugEnabled()) {
        	addLogger( chain );
        }
        
        if (config.isUseSSL()) {
        	addSSLSupport(chain);
        }

        chain.addLast( "codec", new ProtocolCodecFilter(
            	new TCPMsgCodecFactory(config)));
        
        socketAddress = new InetSocketAddress(
            	InetAddress.getByName(this.config.getLocalAddress()),
            	this.config.getLocalPort() );
        serverHandler = new TCPServerProtocolHandler(this);
        
        // Bind
        startAccept();

        log.info( "Listening on port " + this.config.getLocalPort() );
    }
    
    public void stop() throws Exception {
        super.stop();
        if (serverHandler != null) {
        	serverHandler.dispose();
        }
        acceptor.unbindAll();
    }
    
    public void pauseAccept() {
    	acceptor.unbindAll();
    }
    
    public void startAccept() throws IOException {
    	log.debug("Acceptor managing socket? "+acceptor.isManaged(socketAddress));
    	if (!acceptor.isManaged(socketAddress)) {
   			acceptor.bind(socketAddress, serverHandler, acceptorConfig);
    	}
    	log.debug("Acceptor managing socket? "+acceptor.isManaged(socketAddress));
    }

    public void process(MessageExchange exchange) throws Exception {
        // TODO: As we act as a consumer (we just send JBI exchanges)
        // we will receive responses or DONE / ERROR status here
    	
    	log.debug("Received ME: "+exchange);
    	String correlationId = (String)exchange.getProperty(JbiConstants.CORRELATION_ID);
    	if (correlationId == null) {
    		log.warn("CorrelationId null. The client will not receive response");
    		return;
    	}
    	synchronized (correlatedSessions) {
			
	    	IoSession session = correlatedSessions.get(correlationId);
	    	if (session == null) {
	    		throw new IOException("CorrelationID ["+correlationId+"] does not identify any TCP session");
	    	}
	    	if (!session.isConnected()) {
	    		log.warn("The tcp session is not connected The client will not receive response. " +
	    			"CorrelationId="+correlationId);
	    		removeCorrelatedSession(correlationId);
	    		return;
	    	}
	    	
	    	if (config.getDefaultMep().toString().equals(
	    			CommonConfig.DEFINOUTMEP)) {
	    		NormalizedMessage out = getOutMessage(exchange);
	    		
	    		String xpathString = (config.isUseInNMEnvelope())?
	    			config.getInNmEnvelope()+"/text()" :
	    			"text()";
	    		Object content = NMUtils.retrieveMessageInEnvelope(out, xpathString);
//		    	SourceTransformer trans = new SourceTransformer();
//		    	String tcpMsg = trans.contentToString(out);
		    	log.debug("Sending response for correlationId: "+correlationId+
		    			" msg: "+content);
		    	
		    	// terminate the exchange
		    	done(exchange);
		    	if (config.isBase64decode()) {
		    		try {
		    			session.write(Base64.decodeBase64((byte[])content));
		    		} catch (ClassCastException cce) {
		    			log.error("Error writing response", cce);
		    		}
		    	} else {
		    		session.write(content);
		    	}
	    	}
    	
    	}
    }
    
    private NormalizedMessage getOutMessage(MessageExchange exchange)
    	throws NoOutMessageAvailableException {
    	
    	NormalizedMessage in = exchange.getMessage("out");
    	if (in == null) {
    		throw new NoOutMessageAvailableException(exchange);
    	}
    	return in;
    }

    private void addLogger( DefaultIoFilterChainBuilder chain ) throws Exception
    {
    	// we have to log CONNECTION and DATA.
    	// TODO CHECK If the logger is added after the codec it receives data already
    	// converted to the codec protocol?
        chain.addLast( "logger", new LoggingFilter() );
        log.info( "Logging ON" );
    }
    
    private void addSSLSupport( DefaultIoFilterChainBuilder chain )
            throws Exception
    {
        SSLFilter sslFilter = new SSLFilter( TCPBCSSLContextFactory
                .getInstance( config ) );
        sslFilter.setNeedClientAuth(config.isUseSSLClientMode());
        chain.addLast( "sslFilter", sslFilter );
        log.info( "SSL ON" );
    }    

	public TCPBCConfig getConfig() {
		return config;
	}

	public void setConfig(TCPBCConfig config) {
		this.config = config;
	}
	
	public synchronized void addCorrelatedSession(String id, IoSession session) {
		log.debug("Adding correlationID: "+id);
		correlatedSessions.put(id, session);
	}
	
	public synchronized void removeCorrelatedSession(String id) {
		log.debug("Removing correlationID: "+id);
		correlatedSessions.remove(id);
	}
}
