/**

    Copyright 2007 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package it.eng.spagosoa.smx.components.tcp.codec;

import java.nio.ByteOrder;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.mina.common.ByteBuffer;
import org.apache.mina.common.IoSession;
import org.apache.mina.filter.codec.ProtocolDecoder;
import org.apache.mina.filter.codec.ProtocolDecoderOutput;
import org.apache.mina.util.SessionLog;

/**
 * Decode bytes received by the socket into messages for the protocol.
 * Messages are recognize having an 'header' and a 'trailer' that bound them.
 * Data is passed to the ProtocolDecoderOutput as ByteBuffer.
 * 
 * TODO
 * Now data is passed as ByteBuffer but should be encapsulated into objects
 * that rappresent messages.
 * 
 * @author buso
 *
 */
public class TCPMsgDecoder implements ProtocolDecoder {
	private static final String CONTEXT = "TCPMSGDECODER_CONTEXT".intern();
	private static final Log log = LogFactory.getLog(TCPMsgDecoder.class);
	
	private boolean stripped = false;
	private ByteBuffer header;
	private ByteBuffer trailer;
	private ByteOrder order;
	
	public TCPMsgDecoder(ByteBuffer header, ByteBuffer trailer, boolean strip,
		ByteOrder endianness) {
		this.header = header.duplicate();
		this.trailer = trailer.duplicate();
		stripped = strip;
		order = endianness;
		
		this.header.order(order);
		this.trailer.order(order);
	}
	
	/**
	 * Identify messages in the passed ByteBuffer. If a message is not entirely
	 * received via the buffer, the chunk is saved in session and re-decoded in
	 * the next call.
	 * 
	 */
	public void decode(IoSession session, ByteBuffer in,
			ProtocolDecoderOutput out) throws Exception {

		if (log.isDebugEnabled()) {
			log.debug(session.getAttribute(SessionLog.PREFIX)+"decode start");
		}

		Context ctx = getContext(session);
		int headMatchCount = ctx.getHeadMatchCount();
		int trailMatchCount = ctx.getTrailMatchCount();
		
		if (headMatchCount > 0 || trailMatchCount > 0) {
	        log.debug("headMatchCount="+headMatchCount);
	        log.debug("trailMatchCount="+trailMatchCount);
	        log.debug("ctx buffer");
			logBuffer(ctx.getBuffer());
		}
		
		ByteBuffer buffer = ByteBuffer.allocate(1024).setAutoExpand(true).order(order); 
		ByteBuffer msg = ByteBuffer.allocate(1024).setAutoExpand(true).order(order); 
		buffer.put(ctx.getBuffer());
		buffer.flip();

		ByteBuffer internalInBuffer = ByteBuffer.allocate(1024).setAutoExpand(true).order(order);
		internalInBuffer.put(buffer);
		internalInBuffer.put(in);
		internalInBuffer.flip();
		
		ByteBuffer trailerBuffer = ByteBuffer.allocate(trailer.capacity()).order(order);
		// TODO put all in an object that represent the protocol message
		
		headMatchCount = 0;
        boolean messageFinished = false;
        while( internalInBuffer.hasRemaining() ) {
        	messageFinished = false;
            byte b = internalInBuffer.get();
            buffer.put(b);
            if( header.get( headMatchCount ) == b ) {
            	headMatchCount ++;
                if( headMatchCount == header.limit() )
                { // the header terminate
                	if (!stripped) {
                		msg.put(b);
                	}
                	// retrieve the body looking for trailer
                	while (internalInBuffer.hasRemaining() && !messageFinished) {
                		b = internalInBuffer.get();
                		buffer.put(b);
                		if ( trailer.get( trailMatchCount ) == b ) {
                			trailMatchCount++;
                			trailerBuffer.put(b);
                			if (trailMatchCount == trailer.limit()) {
                				// the message is ended

                				if (!stripped) {
                		        	trailerBuffer.flip();
                		        	msg.put(trailerBuffer);
                		        }
                		        msg.flip();
                		        
                				out.write(msg);
                				
                				buffer.release();
//                				buffer = ByteBuffer.allocate(1024).setAutoExpand(true).order(order);
                				msg = ByteBuffer.allocate(1024).setAutoExpand(true).order(order);
                				buffer.position(0);
                				buffer.limit(0);
                				trailerBuffer.clear();
                				messageFinished = true;
                				headMatchCount = 0;
                				trailMatchCount = 0;
                			}
                		} else {
                			// this is part of the body
                			if (trailMatchCount > 0) {
                				// somethings that start as the trailer was found
                				trailerBuffer.flip();
                				msg.put(trailerBuffer);
                				trailerBuffer.clear();
                			}
                			
                			trailMatchCount = 0;
                			msg.put(b);
                		}
                	}
                } else {
                	if (!stripped) {
                		msg.put(b);
                	}
                }
            } else {
            	headMatchCount = 0;
            }
        }
        
        if (headMatchCount > 0 || trailMatchCount > 0) {
            log.debug("headMatchCount="+headMatchCount);
            log.debug("trailMatchCount="+trailMatchCount);
	        log.debug("buffer");
        	logBuffer(buffer);
        }
    	if (!messageFinished) {
    		log.debug("Stop reading, but message is not finished! " +
    				"received messages: "+session.getReadMessages());
    		// the readed buffer is putted in session and re-decoded on next call

    		log.debug("Chunked buffer size (limit): "+buffer.limit());
    	}
		// set the headMatchCount and trailMatchCount on Context
        ctx.setHeadMatchCount(headMatchCount);
        ctx.setTrailMatchCount(trailMatchCount);

        // put remainder on context buffer
        ctx.setBuffer(buffer);
        if (log.isDebugEnabled()) {
			log.debug(session.getAttribute(SessionLog.PREFIX)+"decode finished");
        }
	}
	
	private void logBuffer(ByteBuffer buff) throws Exception {
		int pos = buff.position();
		int limit = buff.limit();
		log.debug("Buffer.toString(): "+buff);
		log.debug("Buffer: "+buff.rewind().getString(java.nio
			.charset.Charset.forName( "UTF-8" ).newDecoder()));
		buff.position(pos);
		buff.limit(limit);
	}
	
	private Context getContext(IoSession session) {
        Context ctx;
        ctx = ( Context ) session.getAttribute( CONTEXT );
        if( ctx == null )
        {
            ctx = new Context();
            session.setAttribute( CONTEXT, ctx );
        }
        return ctx;		
	} 

	public void dispose(IoSession session) throws Exception {
		// TODO Auto-generated method stub

	}

	public void finishDecode(IoSession session, ProtocolDecoderOutput out)
			throws Exception {
		// TODO Auto-generated method stub

	}
	
	/**
	 * Context used to put informations on IoSession
	 * 
	 * @author buso
	 */
	private class Context {
		private ByteBuffer buffer;
		private int headMatchCount = 0;
		private int trailMatchCount = 0;
		
        private Context()
        {
            buffer = ByteBuffer.allocate( 1024 ).setAutoExpand( true ).order(order);
            buffer.limit(buffer.position());
        }

		public ByteBuffer getBuffer() {
			return buffer;
		}
		
		public void setBuffer(ByteBuffer buff) {
			buffer = buff;
			buffer.rewind();
		}

		public int getHeadMatchCount() {
			return headMatchCount;
		}

		public void setHeadMatchCount(int headMatchCount) {
			this.headMatchCount = headMatchCount;
		}

		public int getTrailMatchCount() {
			return trailMatchCount;
		}

		public void setTrailMatchCount(int trailMatchCount) {
			this.trailMatchCount = trailMatchCount;
		}
	}
}
