/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.servicemix.components.jdbc;

import java.sql.Connection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.sql.DataSource;
import javax.xml.transform.Source;

import org.apache.commons.dbutils.DbUtils;
import org.apache.commons.dbutils.QueryRunner;
import org.apache.commons.dbutils.handlers.MapListHandler;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.MessageExchangeListener;
import org.apache.servicemix.components.util.TransformComponentSupport;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.apache.xpath.CachedXPathAPI;
import org.w3c.dom.Node;

public class JdbcComponent extends TransformComponentSupport implements MessageExchangeListener {
    private static final Log log = LogFactory.getLog(JdbcComponent.class);

    private DataSource dataSource;
    private boolean responseRequired = false;

    public boolean transform(MessageExchange exchange, NormalizedMessage in, NormalizedMessage out) throws MessagingException {
        log.debug("Received a JDBC request. Datasource=" + dataSource + ", ResponseRequired=" + responseRequired);
       

        try {

            SourceTransformer domTransform = new SourceTransformer();
            Node domNode = domTransform.toDOMNode(in); 

            String query = getQuery(domNode);
            log.debug("Retrieved query: " + query);
 
            Source outMsg = internalExecute(dataSource, query);
           
            if (outMsg != null) {
                // There is a valid response
                log.debug("Response: " + domTransform.toString(outMsg));
                out.setContent(outMsg);
                return true;

            } else if (responseRequired) {
                // Create an empty <sqlResult> element
                log.debug("Response: Empty Response");
                out.setContent(toXmlSource());
                return true;

            } else {
                log.debug("Response: No Response");
                // There is no valid response
                return false;
            }
        } catch (Exception e) {
            log.error("JDBC Component Exception: ", e);
//            out.setContent(createExceptionXml(e));
//            return true;
            throw new MessagingException(e);
        } 
           
    }

    public Source internalExecute(DataSource ds, String query){
    	Connection conn = null;
    	Source returnedSource = null;
    	try{
    		log.debug("Getting Connection");
    		conn = dataSource.getConnection();
    		QueryRunner run = new QueryRunner();
    		String trimmedQuery = query.trim();
    		log.debug("Query Is "+ trimmedQuery);
    		if (trimmedQuery.startsWith("SELECT") || trimmedQuery.startsWith("select")){
    			log.debug("Query is a SELECT ");
    			List<Map> result = (List<Map>) run.query(conn, trimmedQuery, new MapListHandler());
    			log.debug("Result "+ result.size());
    			returnedSource = toXmlSource(result);
    		}else{
    			log.debug("Query is Insert or Update");
    			int updated = run.update(conn, trimmedQuery);
    			log.debug("Updated ["+updated+"] Records");
    			returnedSource = toXmlSource(updated);
    			
    		}
    		log.debug("Closing Connection");
    		DbUtils.commitAndCloseQuietly(conn);
    		log.debug("Connection Closed");
    	}catch (Throwable e) {
    		log.error(e);
			DbUtils.rollbackAndCloseQuietly(conn);
		}
    	return returnedSource;
    }
    public String getQuery(Node node) throws Exception {
        CachedXPathAPI xpath = new CachedXPathAPI();

        node = xpath.selectSingleNode(node, "sql/child::text()");

        // First child should be <sql></sql> element
        if (node == null) {
            throw new IllegalStateException("Expecting <sql></sql> node. Found: " + node);
        }

        return node.getNodeValue();
    }

    public void setDataSource(DataSource ds) {
        dataSource = ds;
    }

    public DataSource getDataSource() {
        return dataSource;
    }

    /**
     * If true, an empty <sqlResult> element is created and send as a response if there is no result
     * @param val
     */
    public void setResponseRequired(boolean val) {
        responseRequired = val;
    }

    public boolean getResponseRequired() {
        return responseRequired;
    }
/*
    protected Source toXmlSource(ResultSet rs) throws Exception {

        ResultSetMetaData meta = rs.getMetaData();
        int colCount = meta.getColumnCount();

        String[] colNames = getUniqueColumnNames(meta);
        
        StringBuffer buff = new StringBuffer("");

        while (rs.next()) {
            buff.append("<row ");
            for (int i=1; i<=colCount; i++) {
                buff.append(colNames[i].toLowerCase() + "='" + rs.getString(i) + "' ");
            }
            buff.append("/>");
        }

        if (buff.length() > 0) {
            // If non-empty result, insert parent tags
            buff.insert(0, "<sqlResult><resultSet>");
            buff.append("</resultSet></sqlResult>");
        } else {
            // If empty result, return null source
            return null;
        }

        return new StringSource(buff.toString());
    }
*/
    protected Source toXmlSource(List<Map> mapListResult) throws Exception {
    	StringBuffer buff = new StringBuffer();
    	buff.append("<sqlResult>");
    	if (mapListResult.size() > 0){
    		buff.append("<resultSet>");
            Iterator itKey = null;
    		String key = null;
            for (Map m : mapListResult){
    			 buff.append("<row ");
    			 itKey = m.keySet().iterator();
    			 while (itKey.hasNext()){
    				 key = (String)itKey.next();
    				 buff.append(key.toLowerCase() + "='" + String.valueOf(m.get(key)) + "' ");
    			 }
    			 buff.append("/>");
    		}
            buff.append("</resultSet>");
    	}
    	buff.append("</sqlResult>");
        return new StringSource(buff.toString());
    }

    /**
     * Returns a String[] containing unique ColNames. 
   
    protected String[] getUniqueColumnNames(ResultSetMetaData metaData) throws SQLException {

        List colNames = new LinkedList();
        Map chanedNames = new HashMap();

        for (int i = 0; i < metaData.getColumnCount(); i++) {
            String name = metaData.getColumnName(i);

            if (name.equals("")) {
                name = "__unknown_column__";
            }

            if (colNames.contains(name)) {

                int count;
                if (chanedNames.containsKey(name)) {
                    Integer integer = (Integer) chanedNames.get(name);
                    Integer newInteger = new Integer(integer.intValue() + 1);
                    chanedNames.put(name, newInteger);
                    count = newInteger.intValue();
                } else {
                    chanedNames.put(name, new Integer(1));
                    count = 1;
                }

                name = name + "_" + count;

            }

            colNames.add(name);

        }

        return (String[]) colNames.toArray(new String[colNames.size()]);
    }
  */
    protected Source toXmlSource(int updateCount) throws Exception {
        return new StringSource("<sqlResult><updateCount value='" + updateCount + "'/></sqlResult>");
    }

    protected Source toXmlSource() throws Exception {
        return new StringSource("<sqlResult></sqlResult>");
    }


     //Custom method to create and handle exception
    /*
    public boolean isExceptionXml(Node document) {
        return document.getFirstChild().getNodeName().equalsIgnoreCase("exception");
    }
   
    public Source createExceptionXml(Throwable e) {
        return new StringSource("<exception class='" + e.getClass() + "'>'" + e.getMessage() + "'</exception>");
    }
     */ 
}

